package denim.util;

import beast.base.core.Description;
import beast.base.evolution.tree.Node;
import beast.base.evolution.tree.TreeInterface;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/*
 * Created by GRJ on 12/06/2017.
 */

@Description("Helper class for CoalescenceMigrationDistribution. Deals with species tree information")
public class STreeBranchesEpochs {


    private TreeInterface tree;
    private int nNodes;

    private boolean [][] canMigrate;
    // canMigrate[b][d] is true iff b and d can migrate
    private int [] initialEpochs;
    // initialEpochs[b] is the epoch at the start (tipward end) of branch b
    private double [] epochHeights;
    // epochHeights[e] is start of e'th epoch. epochHeights[0] = 0.
    private boolean [][] epochBranchExists;
    // epochBranchExists[e][b] is true iff branch b exists during epoch e.
    private ArrayList<ArrayList<STreeBranchIncident>> branchBoundaries;

    private double [][] tmrcas;
    // tmrcas[b][d] is the time of the most recent ancestor of b and d
    private int [][] topologicalDistances;
    // topologicalDistances[b][d] is the number of branches between b and d





    public STreeBranchesEpochs(TreeInterface tree) {
        this.tree = tree;
        nNodes = tree.getNodeCount();
        int nTips = tree.getLeafNodeCount();

        tmrcas = new double[nNodes][nNodes];
        topologicalDistances = new int[nNodes][nNodes];
        canMigrate = new boolean[nNodes][nNodes];
        initialEpochs = new int[nNodes];
        epochHeights = new double[nTips];
        epochBranchExists = new boolean[nTips][nNodes];
        branchBoundaries = new ArrayList<>();
    }



    // updating methods. update() is always called. updateBranchPairDists() is for multiple migration rates.

    public void update(double maxGtreeHeight) {
        updateEpochHeights();
        updateEpochBranchExists();
        updateCanMigrate();
        updateBranchBoundariesAndInitialEpochs(maxGtreeHeight);
    }



    public void updateBranchPairDists() {
        for (int b = 0; b < nNodes; b++) {
            for (int d = 0; d < nNodes; d++) {
                Node mrca = TreeMethods.mrca(tree.getNode(b), tree.getNode(d));
                tmrcas[b][d] = mrca.getHeight();
                topologicalDistances[b][d] = TreeMethods.distanceBetweenNodes(tree.getNode(b), tree.getNode(d));
            }
        }
    }


    // get methods.

    public int getInitialEpoch(int b) {
        return initialEpochs[b];
    }

    public ArrayList<STreeBranchIncident> getBoundaries(int b) {
        return branchBoundaries.get(b);
    }


    public boolean branchesCanMigrate(int b, int d) {
        return canMigrate[b][d];
    }


    public int epochCount() {
        return epochHeights.length;
    }

    public boolean overlapsInEpoch(int b, int d, int e) {
        return epochBranchExists[e][b]  &&  epochBranchExists[e][d];
    }


    //  get methods  for  multiple migration rates.

    public double getTmrca(int b, int d) {
        return tmrcas[b][d];
    }

    public int getTopologicalDistance(int b, int d) {
        return topologicalDistances[b][d];
    }



/**********************************************************************************************************************/


    private void updateEpochHeights() {
        List<Node> intNodes = tree.getInternalNodes();
        assert intNodes.size() + 1 == epochHeights.length;
        epochHeights[0] = 0.0;
        for (int n = 0; n < intNodes.size(); n++) {
            epochHeights[n+1] = intNodes.get(n).getHeight();
        }
        Arrays.sort(epochHeights);
    }



    private void updateEpochBranchExists() {
        for (int e = 0; e < epochBranchExists.length; e++) {
            for (int b = 0; b < nNodes; b++) {
                epochBranchExists[e][b] = false;
            }
        }
        for (int e = 0; e+1 < epochHeights.length; e++) {
            double h0 = epochHeights[e];
            double h1 = epochHeights[e+1];
            for (int b = 0; b < nNodes; b++) {
                Node nb = tree.getNode(b);
                double bh0 = nb.getHeight();
                if (!nb.isRoot()) {
                    double bh1 = nb.getParent().getHeight();
                    if (bh0 <= h0  &&  bh1 >= h1) {
                        epochBranchExists[e][b] = true;
                    }
                }
            }
        }
    }


    private void updateCanMigrate() {
        for (int b = 0; b < nNodes; b++) {
            for (int d = 0; d < nNodes; d++) {
                if (b == d) {
                    canMigrate[b][d] = false;
                    continue;
                }
                Node nb = tree.getNode(b);
                Node nd = tree.getNode(d);
                if (nb.isRoot() || nd.isRoot()) {
                    canMigrate[b][d] = false;
                    continue;
                }
                double b0 = nb.getHeight();
                double b1 = nb.getParent().getHeight();
                double d0 = nd.getHeight();
                double d1 = nd.getParent().getHeight();
                double t0 = Math.max(b0, d0);
                double t1 = Math.min(b1, d1);
                if (t0 >= t1) {
                    canMigrate[b][d] = false;
                } else {
                    canMigrate[b][d] = true;
                }

            }
        }
    }


    private void updateBranchBoundariesAndInitialEpochs(double maxGtreeHeight) {
        // For each stree branch b, collect incidents for branch start/end, epoch changes.
        // Also, the initial epoch the branch starts in
        // These are the same for all j.
        // Take care to make b correspond to node number in ArrayList.

        branchBoundaries = new ArrayList<>();
        for (int b = 0; b < nNodes; b++) {
            Node nb = tree.getNode(b);
            ArrayList<STreeBranchIncident> boundaries_b = new ArrayList<>();
            double height = nb.getHeight();
            double ancHeight;
            boundaries_b.add(STreeBranchIncident.makeBranchStart(height));
            if (!nb.isRoot()) {
                ancHeight = nb.getParent().getHeight();
            } else {
                ancHeight = maxGtreeHeight;
            }
            initialEpochs[b] = -1;
            for (int e = 0; e < epochHeights.length; e++) {
                if (epochHeights[e] == nb.getHeight()) {
                    initialEpochs[b] = e;
                }
                if (epochHeights[e] > nb.getHeight() && epochHeights[e] < ancHeight) {
                    boundaries_b.add(STreeBranchIncident.makeEpochIncident(epochHeights[e], e));
                }
            }
            assert initialEpochs[b] >= 0;
            boundaries_b.add(STreeBranchIncident.makeBranchEnd(ancHeight));
            branchBoundaries.add(boundaries_b);
        }
    }


}
